@@ -17,6 +17,7 @@ SELECT octet_length(get_raw_page('test1', 'fsm', 1)) AS fsm_1;
 SELECT octet_length(get_raw_page('test1', 'vm', 0)) AS vm_0;
 SELECT octet_length(get_raw_page('test1', 'vm', 1)) AS vm_1;
 
+SELECT octet_length(get_raw_page('test1', 'main', -1));
 SELECT octet_length(get_raw_page('xxx', 'main', 0));
 SELECT octet_length(get_raw_page('test1', 'xxx', 0));
 
@@ -25,6 +26,7 @@ SELECT get_raw_page('test1', 0) = get_raw_page('test1', 'main', 0);
 SELECT pagesize, version FROM page_header(get_raw_page('test1', 0));
 
 SELECT page_checksum(get_raw_page('test1', 0), 0) IS NOT NULL AS silly_checksum_test;
+SELECT page_checksum(get_raw_page('test1', 0), -1);
 
 SELECT tuple_data_split('test1'::regclass, t_data, t_infomask, t_infomask2, t_bits)
     FROM heap_page_items(get_raw_page('test1', 0));
